from .anneal_basic import anneal as anneal_basic
from .anneal_qac import anneal as anneal_qac
from .anneal_copy import anneal as anneal_copy

from .prune_functions import abs_smallest_prune, no_prune


def anneal(config, iter, env, mu):
    """Controls the annealing and energy selection process.

    TODO: split into multiple (two?) functions, depending on regular mode or qac mode

    Called from train() multiple times. Each call to anneal performs a training step
    on the D-Wave hardware.

    Parameters:
    - config       Configuration struct (hold many useful config params)
    - iter         Declares the current iteration within the training method. This is needed
                   as some configuration parameters are per-iteration arrays.
    - env          Env object that determines all data-processing hyperparameters.
    - mu           Vector of expected values of qubit spins.
    """
    if config.anneal["prune_method"] == "abs_smallest":
        config.anneal["prune_method"] = abs_smallest_prune
    elif config.anneal["prune_method"] == "no_prune":
        config.anneal["prune_method"] = no_prune
    else:  # User specified function
        pass

    if config.anneal["anneal_method"] == "basic":
        return anneal_basic(config, iter, env, mu)
    elif config.anneal["anneal_method"] == "qac":
        return anneal_qac(config, iter, env, mu)
    elif config.anneal["anneal_method"] == "copy":
        return anneal_copy(config, iter, env, mu)
    else:  # User specified function
        return config.anneal["anneal_method"](config, iter, env, mu)



import numpy as np


def abs_smallest_prune(J, params_dict):
    """
    J is an adjacency matrix, but because our graph only has at most a single edge between two nodes,
    only the upper triangle of J encodes the edge weights (ie - J[2,3] is referring to the same edge as
    J[3,2] so we only store it once). As such, we need to compute the cutoff on only the upper triangle
    (the nonzero values) of J. Also, np.percentile doesn't take into account magnitude, only value, so
    in order to not lose all negative values, we use the absolute value of J.

    Parameters:
    - J                     Matrix of couplings between qubits
    - cutoff_percentile     Minimum percentile for couplings to NOT be removed
    """
    shape = np.shape(J)
    flat_J = np.ndarray.flatten(J)
    abs_J = np.abs(flat_J)
    triu_abs_J = np.abs(J[np.triu_indices(shape[0], k=1)])
    cutoff_val = np.percentile(triu_abs_J, params_dict["cutoff_percentile"])
    new_J = np.where(abs_J > cutoff_val, flat_J, 0)

    return np.reshape(new_J, shape)


def no_prune(J, params_dict):
    pass



import datetime
import imp
import json
import os
import time
from random import seed

import numpy as np
import qamlz
from sklearn.model_selection import StratifiedShuffleSplit, train_test_split

from roc_excited_states import main as roc_main

script_path = os.path.dirname(os.path.realpath(__file__))


def make_results_json(fail_flag=""):
    filename = "%sanneal_results-%s.json" % (fail_flag, model.start_time)
    destdir = os.path.join(script_path, "qamlz_runs")
    filepath = os.path.join(destdir, filename)
    if not os.path.exists(destdir):
        os.makedirs(destdir)
    json.dump(model.anneal_results, open(filepath, "w"), indent=4)

    return filename


def make_mus_files(fail_flag=""):
    mus_filenames = []
    mus_destdir = None
    for k, v in model.mus_dict.items():
        mus_filename = fail_flag + k
        mus_filenames.append(mus_filename)
        mus_destdir = os.path.join(script_path, "mus/%s" % model.start_time)
        mus_filepath = os.path.join(mus_destdir, mus_filename)
        if not os.path.exists(mus_destdir):
            os.makedirs(mus_destdir)
        np.save(mus_filepath, v)

    return mus_destdir, mus_filenames


# sig = np.loadtxt('sig.csv')
# bkg = np.loadtxt('bkg.csv')
# # X_train = np.vstack((sig[:100, :], bkg[:100, :]))
# # y_train = np.concatenate((np.ones(100), -np.ones(100)))
# full_X_data = np.vstack((sig, bkg))
# full_y_data = np.concatenate((np.ones(np.size(sig, axis=0)), -np.ones(np.size(bkg, axis=0))))
# sss = StratifiedShuffleSplit(n_splits=8, test_size=0.5, random_state=0)
# for train_index, test_index in sss.split(full_X_data, full_y_data):
#     X_train, X_test = full_X_data[train_index], full_X_data[test_index]
#     y_train, y_test = full_y_data[train_index], full_y_data[test_index]
# # np.random.seed(0)
# rand_train_indeces = np.random.choice(np.size(X_train, axis=0), size=5000)
# rand_test_indeces = np.random.choice(np.size(X_test, axis=0), size=5000)

sig_X = np.loadtxt("sig.csv")
bkg_X = np.loadtxt("bkg.csv")
sig_y = np.ones(np.size(sig_X, axis=0))
bkg_y = -np.ones(np.size(bkg_X, axis=0))
sig_X_train, sig_X_test, sig_y_train, sig_y_test = train_test_split(
    sig_X, sig_y, test_size=0.5, shuffle=True, random_state=0
)
bkg_X_train, bkg_X_test, bkg_y_train, bkg_y_test = train_test_split(
    bkg_X, bkg_y, test_size=0.5, shuffle=True, random_state=0
)

X_train, y_train = np.vstack((sig_X_train, bkg_X_train)), np.concatenate(
    (sig_y_train, bkg_y_train)
)

url = "https://cloud.dwavesys.com/sapi/"
token = os.environ["USC_DWAVE_TOKEN"]
# token = os.environ["DWAVE_TOKEN"]

train_sizes = [100, 1000, 5000, 10000, 15000, 20000]
for train_size in train_sizes:
    print("train_size = %d" % train_size)
    np.random.seed(0)
    rand_train_indeces = np.random.choice(np.size(X_train, axis=0), size=train_size)
    env = qamlz.TrainEnv(
        X_train[rand_train_indeces],
        y_train[rand_train_indeces],
        url,
        token,
        fidelity=7,
        dwave_topology="chimera",
    )
    config = qamlz.ModelConfig()
    config.cutoff_percentile = 95
    config.encode_vars = None
    model = qamlz.Model(config, env)
    try:
        model.train()
    except Exception as e:
        print(e)
        make_results_json("FAIL")
        mus_dir, mus_filenames = make_mus_files("FAIL")
        raise e
    make_results_json()
    mus_dir, mus_filenames = make_mus_files()
    time.sleep(5)
    roc_main(
        sig=sig_X_test,
        bkg=bkg_X_test,
        mus_dir=mus_dir,
        mus_filenames=mus_filenames,
        train_size=train_size,
    )



import numpy as np

from .anneal_functions import make_h_J, make_bqm, dwave_connect, unfix


def anneal(config, iter, env, mu):
    """Controls the annealing and energy selection process.

    Called from train() multiple times. Each call to anneal performs a training step
    on the D-Wave hardware.

    Parameters:
    - config       Configuration struct (hold many useful config params)
    - iter         Declares the current iteration within the training method. This is needed
                   as some configuration parameters are per-iteration arrays.
    - env          Env object that determines all data-processing hyperparameters.
    - mu           Vector of expected values of qubit spins.
    """
    print(f"iteration {iter}")
    h, J = make_h_J(env.C_i, env.C_ij, mu, pow(config.zoom_factor, iter))
    J = config.anneal["prune_method"](J, config.anneal["prune_params"])
    bqm, bqm_nx, fixed_dict = make_bqm(h, J)

    if len(fixed_dict) == np.size(h):
        spin_samples = np.zeros(len(fixed_dict))
        for k, v in fixed_dict.items():
            spin_samples[k] = v

        return [
            np.vstack((spin_samples, spin_samples))
        ]  # change to tile, maybe keep same nreads shape

    spin_samples, energies = dwave_connect(config, iter, env.sampler, bqm, bqm_nx)
    print("min energy = %d, max energy = %d" % (np.amin(energies), np.amax(energies)))

    spin_samples = unfix(spin_samples, np.size(h), fixed_dict=fixed_dict)

    if np.size(spin_samples, axis=0) > config.max_states[iter]:
        excited_states_arr = [spin_samples[-config.max_states[iter] :]]
    else:
        excited_states_arr = [spin_samples]

    return excited_states_arr



import datetime

import numpy as np
from sklearn.metrics import accuracy_score

from .anneal import anneal


def total_hamiltonian(mu, s, sigma, C_i, C_ij):
    """
    Derived from Eq. 9 in QAML-Z paper (ZLokapa et al.)
    -> (Dot products of upper triangle)
    """
    ham_1 = np.einsum("i, i", -C_i, s * sigma)
    ham_2 = np.einsum("i, i", np.einsum("ij, j", np.triu(C_ij, k=1), mu), s * sigma)
    ham_3 = np.einsum(
        "i, i", np.einsum("i, ij", s * sigma, np.triu(C_ij, k=1)), s * sigma
    )

    return ham_1 + ham_2 + ham_3


class ModelConfig:
    def __init__(self):
        """
        Configures the hyperparameters for the model. In essence, this controls how
        the model learns.

        TODO:

        Parameters (of the function):
        - n_iterations          Determines the number of repetitions for the training cycle.

        Configs (Hyperpararmeters of the model):
        - zoom_factor           Range(0,1) Controls the zooming rate per iteration. Essentially a
                                training step; larger zoom_factor means faster to train, but
                                slower to settle.
        - anneal_time           Controls the number of microseconds used by D-Wave machines;
                                larger anneal_time means more accurate results, but each run
                                takes more resources (D-Wave provides limited time).
        - flip_probs            Probailities of random spin-flips; used to prevent overtraining.
        - strengths             Scaling of the qubit (h) and coupling (J) weights.
        - max_states            Number of spin-states to record from D-Wave per iteration.
        - num_reads             Number of times to sample the final state from D-Wave
                                (shouldn't be lower than 100 or else the spin-state distribution
                                may not be properly recorded).
        - prune_vars            Method used to remove weak couplings.
        - cutoff_percentile     Minimum percentile for couplings to NOT be removed.
        - encode_vars           Method used to encode qubits for error-correction
                                (encode_qac, encode_copy, or your own).
        - encoding_depth        Defined for NQAC as the number of copies, can be made into
                                something else for other error-correction schemes.
        - gamma                 Range(0,1] Defined for NQAC as the penalty associated with differing
                                spins for encoded-qubits corresponding to the same
                                logical qubit, can be made into something else for other
                                error-correction schemes.
        - decode_vars           Method used to decode qubits from error-correction
                                (decode_qac, decode_copy, or your own).
        """
        self.n_iterations = 8
        self.zoom_factor = 0.5
        self.anneal_time = 5

        self.flip_probs = np.linspace(0.2, 0.01, num=self.n_iterations)

        self.strengths = [3.0, 1.0, 0.5, 0.2] + [0.1] * (self.n_iterations - 4)
        self.max_states = [16, 4] + [1] * (self.n_iterations - 2)
        self.num_reads = 200

        self.anneal = {
            "anneal_method": "basic",
            "anneal_params": {
                "encoding_depth": 3,
                "gamma": 1,
            },
            "prune_method": "abs_smallest",
            "prune_params": {
                "cutoff_percentile": 95,
            },
            "bqm_method": "undirected_single",
            "bqm_params": {},
        }


class Model:
    """
    Contains the model object (the output of the machine learning).

    TODO: finish results dict -> add in method for final test accuracy (AUROC) computation
     - split into multiple files, with logic for picking which file placed into the config class (which
        is also a seperate file)
    """

    def __init__(self, config, env):
        """
        Initializes the model object, all initialization parameters are contained in the config and env objects.

        TODO: add in hyperparameters in ModelConfig

        Parameters:
        - config            Config object that determines all training hyperparameters.
        - env               Env object that determines all data-processing hyperparameters.

        Model Vars:
        - config            Config object that determines all training hyperparameters.
        - env               Env object that determines all data-processing hyperparameters.
        - start_time        Records the time the model began training, useful for data storage.
        - anneal_results    Output dict of simple accuracies (accuracies computed with predict). #einsum
        - mus_dict          Output dict of optimized weightings (the actual classifier trained).
        """
        self.config = config
        self.env = env

        self.start_time = None
        self.anneal_results = {}
        self.mus_dict = {}

    def train(self):
        """
        Performs the training of the ML model.

        TODO: Finish support for multiclass (parallel One vs. Rest)
        """

        print(f'{"NEW TRAINING":-^30}')
        self.start_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        mus = [np.zeros(np.size(self.env.C_i))]

        for iter in range(self.config.n_iterations):
            new_mus_arr = []
            for mu in mus:
                excited_states_arr = anneal(self.config, iter, self.env, mu)
                for excited_states in excited_states_arr:
                    new_mus_arr.append(
                        self.pick_excited_states(iter, excited_states, mu)
                    )

            avg_accuracies = np.zeros(len(new_mus_arr))
            idx = 0
            for new_mus in new_mus_arr:
                accuracies = []
                for new_mu in new_mus:
                    y_pred = np.einsum("ij, j", self.env.c_i_val, new_mu)
                    accuracies.append(accuracy_score(self.env.y_val, np.sign(y_pred)))

                if len(accuracies) != 0:
                    avg_accuracies[idx] = np.mean(np.array(accuracies))
                idx += 1

            chosen_new_mus = np.argmax(avg_accuracies)
            self.anneal_results["iter%d_val" % iter] = avg_accuracies[chosen_new_mus]
            print("max val accuracy: %.4f" % avg_accuracies[chosen_new_mus])
            mus = new_mus_arr[chosen_new_mus]
            mus_filename = "mus%05d-%d_iter%d_run%d__%s" % (
                self.env.train_size,
                self.env.fidelity,
                iter,
                chosen_new_mus,
                self.start_time,
            )
            self.mus_dict[mus_filename] = mus

    def pick_excited_states(self, iter, excited_states, mu):
        """
        Selects spin states to keep dependent upon their associated energy.

        TODO: Separate from class to allow user changes? What's even the use of this function..

        Parameters:
        - iter              Current iteration of the training.
        - excited_states    Obtained spins of a given read from D-Wave.
        - mu                Spins of qubits for this iteration.
        """
        new_mus = []
        for excited_state in excited_states:
            new_sigma = pow(self.config.zoom_factor, iter + 1)
            energy_flips = np.ones(np.size(excited_state))
            dwave_shift_energy = total_hamiltonian(
                mu, excited_state, new_sigma, self.env.C_i, self.env.C_ij
            )

            for idx in range(np.size(excited_state)):
                no_shift_state = np.copy(excited_state)
                no_shift_state[idx] = 0
                no_shift_energy = total_hamiltonian(
                    mu, no_shift_state, new_sigma, self.env.C_i, self.env.C_ij
                )
                anti_shift_state = np.copy(excited_state)
                anti_shift_state[idx] *= -1
                anti_shift_energy = total_hamiltonian(
                    mu, anti_shift_state, new_sigma, self.env.C_i, self.env.C_ij
                )

                if dwave_shift_energy < anti_shift_energy:
                    if dwave_shift_energy < no_shift_energy:
                        continue
                    else:
                        energy_flips[idx] = 0
                else:
                    if anti_shift_energy < no_shift_energy:
                        energy_flips[idx] = -1
                    else:
                        energy_flips[idx] = 0

            flip_prob = self.config.flip_probs[iter]
            overtrain_flips = np.random.choice(
                [-1, 1], size=np.size(energy_flips), p=[flip_prob, 1 - flip_prob]
            )
            new_state = excited_state * energy_flips * overtrain_flips
            new_mus.append(mu + new_state * new_sigma)

        return new_mus



import dimod
import dwave as dw
import minorminer
import networkx as nx
import numpy as np


def make_h_J(C_i, C_ij, mu, sigma):
    """
    Creates the node (h) and edge (J) weightings for use in building the problem graph.

    Parameters:
    - C_i           Intermediate data representation needed for D-Wave's BQM format.
    - C_ij          Intermediate data representation needed for D-Wave's BQM format.
    - mu            Spins of qubits for this iteration.
    - sigma         zoom_factor to the power of the iteration.
    """
    h = 2 * sigma * (np.einsum("ij, j", C_ij, mu) - C_i)
    J = 2 * np.triu(C_ij, k=1) * pow(sigma, 2)

    return h, J


def make_bqm(h, J):
    """
    Creates the BQM and associated nx graph. The two differ in that fixed nodes
    are not removed from the nx graph.

    TODO: Clean up code and add in script tags for Graphs

    Parameters:
    - h                     Array of qubit-spin weights.
    - J                     Matrix of coupling strengths between qubits.
    """
    bqm_nx = nx.from_numpy_array(J)

    node_dict, attr_dict = {}, {}
    for node in bqm_nx:
        attr_dict["h_bias"] = h[node]
        node_dict[node] = attr_dict
    nx.set_node_attributes(bqm_nx, node_dict)

    bqm = dimod.from_networkx_graph(
        bqm_nx,
        vartype="SPIN",
        node_attribute_name="h_bias",
        edge_attribute_name="weight",
    )

    lowerE, fixed_dict = dw.preprocessing.lower_bounds.roof_duality(bqm, strict=True)
    if fixed_dict == {} or len(fixed_dict) < (np.size(h) // 2):
        print("loosely fixed")
        lowerE, fixed_dict = dw.preprocessing.lower_bounds.roof_duality(
            bqm, strict=False
        )
    for i in fixed_dict.keys():
        bqm.fix_variable(i, fixed_dict[i])
        bqm_nx.remove_node(i)

    print(f"num nodes = {bqm.num_variables} \n num edges = {bqm.num_interactions}")

    return bqm, bqm_nx, fixed_dict


def scale_weights(th, tJ, strength):
    """
    Scales the weightings dependent upon the iteration. It is unclear
    what purpose this serves.

    Parameters:
    - th            Embedded h array.
    - tJ            Embedded J matrix.
    - strength      Current-iteration-indexed strengths array.
    """
    for k in list(th.keys()):
        th[k] /= strength
    for k in list(tJ.keys()):
        tJ[k] /= strength

    return th, tJ


def dwave_connect(config, iter, sampler, bqm, bqm_nx):
    """
    Handles all the connection to D-Wave, including:
    - Embedding the problem into a D-Wave architecture
    - Sending the problem to be annealed
    - Receiving the results from the annealing
    - Unembedding the problem from a D-Wave architecture
    - Unpacking the results into the obtained spins and their energies

    TODO: Multiply by a before and after embedding

    Parameters:
    - config            Config object that determines all training hyperparameters.
    - iter              Current iteration of the training.
    - sampler
    - bqm
    - bqm_nx
    - anneal_time
    """

    a = np.sign(np.random.rand(bqm.num_variables) - 0.5)

    embedding = minorminer.find_embedding(bqm_nx, sampler.to_networkx_graph())

    h_dict = nx.classes.function.get_node_attributes(bqm_nx, "h_bias")
    J_dict = nx.classes.function.get_edge_attributes(bqm_nx, "weight")
    # Unsure what the commented-out code is supposed to do
    # for k in h_dict:
    #     h_dict[k] *= a[k]
    # for u, v in J_dict:
    #     J_dict[(u, v)] *= a[u] * a[v]

    th, tJ = dw.embedding.embed_ising(
        h_dict,
        J_dict,
        embedding,
        sampler.adjacency,
    )
    th, tJ = scale_weights(th, tJ, config.strengths[iter])

    qaresult = sampler.sample_ising(
        th,
        tJ,
        num_reads=config.num_reads,
        annealing_time=config.anneal_time,
        answer_mode="histogram",
    )
    unembed_qaresult = dw.embedding.unembed_sampleset(qaresult, embedding, bqm)

    a = np.tile(a, (np.size(unembed_qaresult.record.sample, axis=0), 1))
    samples, energies = (
        unembed_qaresult.record.sample * a,
        unembed_qaresult.record.energy,
    )

    return samples, energies


def unfix(spin_samples, h_len, fixed_dict):
    """
    Unpacks the qubits fixed using fix_vars.

    Parameters:
    - spin_samples          Obtained spins of all reads from D-Wave.
    - h_len                 Number of nodes in the graph sent to D-Wave.
    - fixed_dict            Dictionary of fixed nodes (fixed just prior to sending
                            D-Wave the graph in the dwave_connect function).
    """
    spin_samples = np.zeros((np.size(spin_samples, axis=0), h_len))
    j = 0
    for i in range(h_len):
        if i in fixed_dict:
            spin_samples[:, i] = fixed_dict[i]
        else:
            spin_samples[:, i] = spin_samples[:, j]
            j += 1

    return spin_samples



import numpy as np

from .anneal_functions import make_h_J, make_bqm, dwave_connect, unfix


def make_bqm_copy(h, J, C):
    """
    Encodes the logical qubits with the copy method.

    Parameters:
    - h                     Array of qubit-spin weights.
    - J                     Matrix of coupling strengths between qubits.
    - C                     Same as encoding_depth: Defined for NQAC as the number of copies, can be made into
                            something else for other error-correction schemes.
    """
    rows, cols = np.shape(J)
    cp_J = np.zeros((C * rows, C * cols))
    for i in np.arange(C):
        cp_J[i * rows : (i + 1) * rows, i * cols : (i + 1) * cols] = J
    h = np.repeat(h, C)

    return make_bqm(h, cp_J)


def decode_copy(spin_samples, h_len, orig_len, fixed_dict):
    """
    Decodes the qubits encoded with the copy method.

    Parameters:
    - spin_samples          Obtained spins of all reads from D-Wave.
    - enc_h_len             Number of nodes in the graph sent to D-Wave (after copying).
    - orig_len              Number of nodes originally in the graph (before NQAC encoding).
    - fixed_dict            Dictionary of fixed nodes (fixed just prior to sending
                            D-Wave the graph in the dwave_connect function).
    """
    spin_samples = unfix(spin_samples, h_len, fixed_dict)

    multi_spins = np.zeros((np.size(spin_samples, axis=1) // orig_len))
    for i in range(np.size(multi_spins)):
        multi_spins[i] = spin_samples[:, i * orig_len : (i + 1) * orig_len]

    return multi_spins


def anneal(config, iter, env, mu):
    """Controls the annealing and energy selection process.

    TODO: line 151, change to tile?

    Called from train() multiple times. Each call to anneal performs a training step
    on the D-Wave hardware.

    Parameters:
    - config       Configuration struct (hold many useful config params)
    - iter         Declares the current iteration within the training method. This is needed
                   as some configuration parameters are per-iteration arrays.
    - env          Env object that determines all data-processing hyperparameters.
    - mu           Vector of expected values of qubit spins.
    """
    print(f"iteration is {iter}")
    h, J = make_h_J(env.C_i, env.C_ij, mu, pow(config.zoom_factor, iter))
    J = config.anneal["prune_method"](J, config.anneal["prune_params"])

    orig_len = np.size(h)
    bqm, bqm_nx, fixed_dict = make_bqm_copy(
        h, J, config.anneal["anneal_params"]["encoding_depth"]
    )

    if len(fixed_dict) == config.anneal["anneal_params"]["encoding_depth"] * orig_len:
        enc_spin_samples = np.zeros(len(fixed_dict))
        for k, v in fixed_dict.items():
            enc_spin_samples[k] = v

        sub_spin_samples = np.vstack(
            np.split(enc_spin_samples, config.encoding_depth, axis=0)
        )
        spin_samples = np.zeros(orig_len)
        for i in range(orig_len):
            spin_samples[i] = np.sign(np.sum(sub_spin_samples[:, i]))
            if spin_samples[i] == 0:
                spin_samples[i] = np.random.choice([-1, 1])

        return [
            np.vstack((spin_samples, spin_samples))
        ]  # change to tile, maybe keep same nreads shape

    spin_samples, energies = dwave_connect(
        config, iter, env.sampler, bqm, bqm_nx, config.anneal_time
    )

    spin_samples = decode_copy(
        spin_samples,
        config.anneal["anneal_params"]["encoding_depth"] * orig_len,
        orig_len,
        fixed_dict,
    )

    print("min energy = %d, max energy = %d" % (np.amin(energies), np.amax(energies)))

    if np.size(spin_samples, axis=0) > config.max_states[iter]:
        excited_states_arr = [
            np.vstack((spin_samples[0, :], spin_samples[-config.max_states[iter] :, :]))
        ]
    else:
        excited_states_arr = [spin_samples]

    return excited_states_arr
